# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from .models import SMSDevice
from .signals import user_created


class SMSBackend(ModelBackend):
    def __init__(self):
        super(SMSBackend, self).__init__()
        self.UserModel = get_user_model()

    def authenticate(self, number=None, token=None, request=None):
        user = None
        if request:
            device = SMSDevice.get(request, number)
            if device and device.verify_token(token):
                try:
                    user = self.UserModel._default_manager.get_by_natural_key(device.number)
                except self.UserModel.DoesNotExist:
                    user = self.create_user(request, device)

        return user

    def create_user(self, request, device):
        user = None
        if getattr(request, 'otp_sms_create_user', False):
            user = self.UserModel.objects.create_user(device.number)
            user_created.send(sender=user.__class__, request=request, user=user, phone=device.number)
        return user


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import time
from binascii import unhexlify
from django.db import models
from django.utils.module_loading import import_string
from django_otp.oath import TOTP
from django_otp.util import random_hex, hex_validator
from django.utils.translation import ugettext_lazy as _
from phonenumber_field.modelfields import PhoneNumberField
from .conf import settings


def default_key():
    return random_hex(20)


def key_validator(value):
    return hex_validator(20)(value)


class SMSDevice(models.Model):
    number = PhoneNumberField(
        verbose_name=_("Phone number"),
        help_text=_("The mobile number to deliver tokens to.")
    )

    key = models.CharField(
        max_length=40,
        validators=[key_validator],
        default=default_key,
        help_text=_("A random key used to generate tokens (hex-encoded).")
    )

    last_t = models.BigIntegerField(
        default=-1,
        help_text=_("The t value of the latest verified token. The next token must be at a higher time step.")
    )

    class Meta(object):
        verbose_name = _("SMS Device")

    def __unicode__(self):
        return '%s' % self.number

    @classmethod
    def get(cls, request, number):
        device_id = request.session.get(settings.OTP_SMS_SESSION_KEY_DEVICE_ID)
        device = None
        if device_id:
            try:
                device_id = int(device_id)
            except ValueError:
                device_id = None
            try:
                device = cls._default_manager.get(pk=device_id, number=number)
            except cls.DoesNotExist:
                pass
        return device

    @property
    def bin_key(self):
        return unhexlify(self.key.encode())

    def generate_token(self):
        totp = self.totp_obj()
        token = format(totp.token(), '06d')
        message = settings.OTP_SMS_TOKEN_TEMPLATE.format(token=token)
        self._deliver_token(message)

    def _deliver_token(self, token):
        adapter_class = import_string(settings.OTP_SMS_ADAPTER)
        adapter = adapter_class(settings.OTP_SMS_AUTH)
        adapter.send(self.number, token, sender=settings.OTP_SMS_FROM)

    def verify_token(self, token):
        try:
            token = int(token)
        except ValueError:
            verified = False
        else:
            totp = self.totp_obj()
            tolerance = settings.OTP_SMS_TOKEN_VALIDITY

            for offset in range(-tolerance, 1):
                totp.drift = offset
                if (totp.t() > self.last_t) and (totp.token() == token):
                    self.last_t = totp.t()
                    self.save()

                    verified = True
                    break
            else:
                verified = False

        return verified

    def totp_obj(self):
        totp = TOTP(self.bin_key, step=1)
        totp.time = time.time()

        return totp


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from datetime import timedelta
from django.conf import settings as django_settings


class Settings(object):
    _defaults = {
        'OTP_SMS_SESSION_KEY_DEVICE_ID': 'OTP-DEVICE-ID',
        'OTP_SMS_SESSION_KEY_LAST_ATTEMPT_TIME': 'OTP-LAST-TIME',
        'OTP_SMS_SESSION_KEY_ATTEMPT': 'OTP-ATTEMPT',
        'OTP_SMS_COUNT_ATTEMPTS': 3,
        'OTP_SMS_LATENCY_ATTEMPTS': timedelta(minutes=5),
        'OTP_SMS_AUTH': None,
        'OTP_SMS_FROM': None,
        'OTP_SMS_TOKEN_TEMPLATE': '{token}',
        'OTP_SMS_TOKEN_VALIDITY': 300,
        'OTP_SMS_TEST_NUMBER': '+79000000000',
        'OTP_SMS_ADAPTER': 'otp_sms.adapters.ConsoleAdapter',
        'OTP_SMS_NOTIFY_ADMINS_ADAPTER_ERROR': True
    }

    def __getattr__(self, name):
        if hasattr(django_settings, name):
            value = getattr(django_settings, name)
        elif name in self._defaults:
            value = self._defaults[name]
        else:
            raise AttributeError(name)

        return value


settings = Settings()


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from django.test import TestCase
from .models import SMSDevice
from .conf import settings


class SMSDeviceTestCase(TestCase):
    device = None

    def setUp(self):
        self.device = SMSDevice.objects.create(number=settings.OTP_SMS_TEST_NUMBER)

    def test_generate_challenge(self):
        self.device.generate_challenge()


# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-11-14 13:50
from __future__ import unicode_literals

from django.db import migrations
import phonenumber_field.modelfields


class Migration(migrations.Migration):

    dependencies = [
        ('otp_sms', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='smsdevice',
            name='number',
            field=phonenumber_field.modelfields.PhoneNumberField(help_text='The mobile number to deliver tokens to.', max_length=128, verbose_name='Phone number'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import urllib2
import urllib
from .base import BaseAdapter, AdapterError


class ProstorAdapter(BaseAdapter):
    base_url = 'http://api.prostor-sms.ru/messages/v2'

    def __init__(self, auth):
        self.login = auth['LOGIN']
        self.password = auth['PASSWORD']

    def _send_request(self, uri, params=None):
        url = self._get_url(uri, params)
        request = urllib2.Request(url)
        passman = urllib2.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, url, self.login, self.password)
        authhandler = urllib2.HTTPBasicAuthHandler(passman)
        try:
            opener = urllib2.build_opener(authhandler)
            data = opener.open(request).read()
            return data
        except IOError, e:
            raise AdapterError(e.code)

    def _get_url(self, uri, params=None):
        url = "%s/%s/" % (self.base_url, uri)
        param_str = ''
        if params is not None:
            for k, v in params.items():
                if v is None:
                    del params[k]
                if isinstance(v, unicode):
                    params[k] = v.encode('utf-8')
            param_str = urllib.urlencode(params)
        if param_str:
            return "%s?%s" % (url, param_str)
        return url

    def send(self, phone, text, sender=None, status_queue_name=None, schedule_time=None, wap_url=None):
        """Sending sms """
        senders = self.senders()
        if not sender and senders:
            sender = senders[0]
        elif sender not in senders:
            raise AdapterError('Sender %s is not available' % sender)

        params = {'phone': phone, 'text': text, 'sender': sender, 'statusQueueName': status_queue_name,
                  'scheduleTime': schedule_time, 'wapurl': wap_url}
        answer = self._send_request('send', params)

        if not answer.startswith('accepted'):
            raise AdapterError(answer)

        return answer

    def status(self, id):
        """Retrieve sms status by it's id """
        params = {'id': id}
        return self._send_request('status', params)

    def status_queue(self, status_queue_name, limit=5):
        """Retrieve latest statuses from queue """
        params = {'statusQueueName': status_queue_name, 'limit': limit}
        return self._send_request('statusQueue', params)

    def credits(self):
        """Retrieve current credit balance """
        return self._send_request('credits')

    def senders(self):
        """Retrieve available signs """
        answer = self._send_request('senders')
        lines = answer.splitlines()
        senders = []
        for line in lines:
            name, status = line.split(';')[:2]
            if status in ('active', 'default'):
                senders.append(name)
        return senders


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from django.conf import settings
from django.http import HttpResponseRedirect
from django.shortcuts import resolve_url
from django.utils.http import is_safe_url
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.contrib.auth import (
    login as auth_login,
    REDIRECT_FIELD_NAME
)
from formtools.wizard.views import SessionWizardView
from .forms import SMSSendForm, SMSAuthenticationForm
from .signals import user_logged_in


class SMSAuthenticationWizardView(SessionWizardView):
    SMS_FORM_KEY = 'sms'
    AUTH_FORM_KEY = 'auth'
    redirect_field_name = REDIRECT_FIELD_NAME
    redirect_to = None
    form_list = [
        (SMS_FORM_KEY, SMSSendForm),
        (AUTH_FORM_KEY, SMSAuthenticationForm),
    ]
    create_user_if_not_exists = False

    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        request.otp_sms_create_user = self.create_user_if_not_exists
        return super(SMSAuthenticationWizardView, self).dispatch(request, *args, **kwargs)

    def get_form_kwargs(self, step=None):
        kwargs = super(SMSAuthenticationWizardView, self).get_form_kwargs(step)
        kwargs.update({
            'request': self.request
        })
        return kwargs

    def get_initial_for_auth_form(self):
        sms_form_data = self.storage.get_step_data(self.SMS_FORM_KEY) or {}
        return {
            'username': sms_form_data.get('%s-number' % self.SMS_FORM_KEY)
        }

    def get_form_initial(self, step):
        initial = super(SMSAuthenticationWizardView, self).get_form_initial(step)
        if step == self.AUTH_FORM_KEY:
            initial.update(self.get_initial_for_auth_form())
        return initial

    def render_next_step(self, form, **kwargs):
        if self.steps.current == self.SMS_FORM_KEY:
            form.save()
            if form.errors:
                return self.render(form)
        return super(SMSAuthenticationWizardView, self).render_next_step(form, **kwargs)

    def render_done(self, form, **kwargs):
        user = form.get_user()
        phone = form.cleaned_data.get('username')
        auth_login(self.request, user)
        user_logged_in.send(sender=user.__class__, request=self.request, user=user, phone=phone)
        done_response = HttpResponseRedirect(self.get_redirect_to())
        self.storage.reset()
        return done_response

    def done(self, form_list, **kwargs):
        pass

    def get_redirect_to(self):
        if self.redirect_to:
            return self.redirect_to

        redirect_to = self.request.POST.get(self.redirect_field_name,
                                            self.request.GET.get(self.redirect_field_name, ''))

        if not is_safe_url(url=redirect_to, host=self.request.get_host()):
            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)

        return redirect_to


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

from django.dispatch import Signal

user_logged_in = Signal(providing_args=["request", "user", "phone"])
user_created = Signal(providing_args=["request", "user", "phone"])


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import urllib2
import urllib
import hashlib
import time
from django.utils.translation import ugettext_lazy as _
from .base import BaseAdapter, AdapterError


class SMSRuAdapter(BaseAdapter):
    BASE_URL = 'https://sms.ru'

    ANSWER_STATUS = {
        -1: _("Message not found"),
        100: _("Message is in the queue"),
        101: _("Message is on the way to the operator"),
        102: _("Message is on the way to the recipient"),
        103: _("Message delivered"),
        104: _("Message failed: out of time"),
        105: _("Message failed: cancelled by the operator"),
        106: _("Message failed: phone malfunction"),
        107: _("Message failed: unknown reason"),
        108: _("Message declined"),
        130: _("Message failed: exceeded limit of messages to this number per day"),
        131: _("Message failed: exceeded limit of identical messages to this number per minute"),
        132: _("Message failed: exceeded limit of identical messages to this number per day"),

        200: _("The supplied API key is wrong"),
        201: _("Out of money"),
        202: _("Bad recipient"),
        203: _("Message text not specified"),
        204: _("Bad sender (unapproved)"),
        205: _("Message too long"),
        206: _("Day message limit reached"),
        207: _("Can't send messages to that number"),
        208: _("Wrong time"),
        209: _("Blacklisted recipient"),
        210: _("GET used when POST must have been"),
        211: _("Unknown method"),
        212: _("Text must be in UTF-8"),
        220: _("The service is temporarily unavailable"),
        230: _("Exceeded limit of messages per day"),
        231: _("Exceeded limit of identical messages per minute"),
        232: _("Exceeded limit of identical messages per day"),
        300: _("Invalid token"),
        301: _("Wrong password"),
        302: _("User authenticated, but not confirmed"),
    }

    def __init__(self, auth):
        super(SMSRuAdapter, self).__init__(auth)
        self._token = None
        self._token_ts = 0
        self.login = auth.get('LOGIN')
        self.password = auth.get('PASSWORD')
        self.key = auth['KEY']

    def _call_raw(self, method, args={}):
        """Calls a remote method."""
        if not isinstance(args, dict):
            raise ValueError("args must be a dictionary")
        args["api_id"] = self.key

        if method in ("sms/send", "sms/cost") and self.login and self.password:
            args["login"] = self.login
            args["token"] = self._get_token()
            args["sig"] = hashlib.md5(self.password + args["token"]).hexdigest()
            del args["api_id"]

        url = "%s/%s?%s" % (self.BASE_URL, method, urllib.urlencode(args))
        try:
            res = urllib2.urlopen(url).read()
        except IOError as e:
            raise AdapterError(e.code)
        return res

    def _call(self, method, args={}):
        res = self._call_raw(method, args)
        res = res.strip().split("\n")

        code, params = res[0], res[1:]
        code = int(code)

        if code not in self.ANSWER_STATUS:
            raise AdapterError(_("Unknown status"))

        if code >= 200:
            raise AdapterError(self.ANSWER_STATUS[code])

        return code, params

    def _get_token(self):
        """Returns a token.  Refreshes it if necessary."""
        if self._token_ts < time.time() - 500:
            self._token = None
        if self._token is None:
            self._token = self.token()
            self._token_ts = time.time()
        return self._token

    def send(self, to, text, sender=None, translit=False, test=False):
        """Sends the message to the specified recipient.  Returns a numeric
        status code, its text description and, if the message was successfully
        accepted, its reference number."""
        if not isinstance(text, unicode):
            text = unicode(text)
        args = {"to": to, "text": text.encode("utf-8")}
        if sender:
            args["from"] = sender
        if translit:
            args["translit"] = "1"
        if test:
            args["test"] = "1"

        code, message_ids = self._call("sms/send", args)
        return message_ids[0]

    def status(self, msgid):
        """Returns message status."""
        code, params = self._call("sms/status", {"id": msgid})
        return self.STATUS_STATUS[code]

    def cost(self, to, message):
        """Prints the cost of the message."""
        code, (cost, length) = self._call("sms/cost", {"to": to, "text": message.encode("utf-8")})
        return cost, length

    def balance(self):
        """Returns your current balance."""
        code, (balance,) = self._call("my/balance")
        return float(balance)

    def limit(self):
        """Returns the remaining message limit."""
        code, (day_limit, sended) = self._call("my/limit")
        return int(day_limit) - int(sended)

    def token(self):
        """Returns a token."""
        return self._call_raw("auth/get_token")


# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import

import time
from datetime import datetime
from django import forms
from django.utils.translation import ugettext_lazy as _
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth import authenticate
from django.core.mail import mail_admins
from phonenumber_field.formfields import PhoneNumberField
from .models import SMSDevice
from .conf import settings
from .adapters import AdapterError


class SMSFormMixin(object):
    def clean_otp(self):
        now = datetime.now()
        attempt = self.request.session.get(settings.OTP_SMS_SESSION_KEY_ATTEMPT, 0)
        last_attempt_time = self.request.session.get(settings.OTP_SMS_SESSION_KEY_LAST_ATTEMPT_TIME)
        if last_attempt_time:
            try:
                last_attempt_time = datetime.fromtimestamp(last_attempt_time)
            except Exception:
                last_attempt_time = None

        if attempt:
            try:
                attempt = int(attempt)
            except ValueError:
                attempt = 0

        if settings.OTP_SMS_COUNT_ATTEMPTS and attempt >= settings.OTP_SMS_COUNT_ATTEMPTS:
            if settings.OTP_SMS_LATENCY_ATTEMPTS is not None and last_attempt_time \
                    and now - last_attempt_time < settings.OTP_SMS_LATENCY_ATTEMPTS:
                raise forms.ValidationError(
                    _('You have exceeded the limit SMS, try after %(minutes)d minutes'),
                    params={'minutes': settings.OTP_SMS_LATENCY_ATTEMPTS.total_seconds() / 60}
                )
            else:
                attempt = 0

        self.request.session[settings.OTP_SMS_SESSION_KEY_LAST_ATTEMPT_TIME] = time.mktime(datetime.now().timetuple())
        self.request.session[settings.OTP_SMS_SESSION_KEY_ATTEMPT] = attempt + 1

    def generate_token(self, device):
        self.request.session[settings.OTP_SMS_SESSION_KEY_DEVICE_ID] = device.pk

        try:
            device.generate_token()
        except AdapterError as e:
            if settings.DEBUG:
                raise e
            elif settings.OTP_SMS_NOTIFY_ADMINS_ADAPTER_ERROR:
                mail_admins('SMS send error', 'AdapterError: %s' % e.message)
            raise forms.ValidationError(_('Error sending sms'))

    def verify_token(self, number, token):
        device = SMSDevice.get(self.request, number)
        return device and device.verify_token(token)


class SMSAuthenticationForm(AuthenticationForm):
    username = PhoneNumberField(label=_('Phone number'))
    password = forms.CharField(label=_('Confirmation code'), widget=forms.PasswordInput)

    error_messages = {
        'invalid_login': _("Please enter a correct phone number and confirmation code"),
        'inactive': _("This account is inactive."),
    }

    def clean(self):
        username = self.cleaned_data.get('username')
        password = self.cleaned_data.get('password')

        if username and password:
            self.user_cache = authenticate(number=username, token=password, request=self.request)
            if self.user_cache is None:
                raise forms.ValidationError(
                    self.error_messages['invalid_login'],
                    code='invalid_login',
                    params={'username': self.username_field.verbose_name},
                )
            else:
                self.confirm_login_allowed(self.user_cache)

        return self.cleaned_data


class SMSSendForm(SMSFormMixin, forms.ModelForm):
    class Meta:
        model = SMSDevice
        fields = ('number',)

    def __init__(self, request, *args, **kwargs):
        self.request = request
        super(SMSSendForm, self).__init__(*args, **kwargs)

    def clean(self):
        self.clean_otp()
        return self.cleaned_data

    def save(self, *args, **kwargs):
        instance = super(SMSSendForm, self).save(*args, **kwargs)

        try:
            self.generate_token(instance)
        except forms.ValidationError as e:
            self.add_error(None, e)
            return None

        return instance


# -*- coding: utf-8 -*-


class BaseAdapter(object):
    def __init__(self, auth):
        self.auth = auth

    def send(self, number, message, **kwargs):
        raise NotImplementedError


class ConsoleAdapter(BaseAdapter):
    def send(self, number, message, **kwargs):
        print
        print 'Sent to %s message "%s"' % (number, message)
        print


class AdapterError(Exception):
    pass


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import phonenumber_field.modelfields
import otp_sms.models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SMSDevice',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('number', phonenumber_field.modelfields.PhoneNumberField(help_text=b'The mobile number to deliver tokens to.', max_length=128)),
                ('key', models.CharField(default=otp_sms.models.default_key, help_text=b'A random key used to generate tokens (hex-encoded).', max_length=40, validators=[otp_sms.models.key_validator])),
                ('last_t', models.BigIntegerField(default=-1, help_text=b'The t value of the latest verified token. The next token must be at a higher time step.')),
            ],
            options={
                'verbose_name': 'SMS Device',
            },
        ),
    ]
